#!/usr/bin/env python

""" === Dwaves.de Python based basic standalone WebServer (DPyWebSrv or DPWS) v1.1 === """
""" uses /usr/lib/python3.11/wsgiref/simple_server.py """
""" licence: GPL 2.0 """
""" what does it do? """
""" as any other webserver it will return *.html, *.css, *.txt, *.jpg and other files hosted in a directory on a server """
""" as any other webserver it can process requests (user pushed a button in a form) """
""" usage:  /path/to/python3 index.py -> will include webserver_wsgi.py, which will then run permanently on port 127.0.0.1:8021 (port can be modified in webserver_wsgi.py)
    then start webbrowser and go to 127.0.0.1:8021, the index.html should be returned :)
"""
import os
import sys
import json
import subprocess
import urllib.parse
import webserver_wsgi
import wsgiref.simple_server

# settings, overwrite settings in index.py
# import the webserver
SETTING_PORT = 8021
# SETTING_PORT = 80 # is possible if run as root
SETTINGS_WEBROOT = "."
forms_data = []  # all submissions

# detect mime type by filename extension
# py will be treated like html, but instead of reading the file, the file will be run and it's output returned to browser
MIME_TYPES = {
            ".gz": "application/gzip" , ".tgz": "application/gzip",
            ".htm": "text/html", ".html": "text/html", ".py": "text/html",
            ".js": "text/javascript",
            ".css": "text/css",
            ".csv": "text/csv",
            ".txt": "text/plain",
            ".ico": "image/x-icon",
            ".jpg": "image/jpeg",
            ".png": "image/png",
            }

def application(environ, start_response):
    PATH_INFO = environ["PATH_INFO"] # requested ./path/to/file.html
    method = environ["REQUEST_METHOD"] # requested method, POST or GET
    request = environ['QUERY_STRING'] # data send via request

    os.chdir(SETTINGS_WEBROOT) # change into webroot dir
    
    if PATH_INFO == "/":              # localhost/ is short for localhost/index.html 
        PATH_INFO = "index.py"

    filename, file_extension = os.path.splitext(PATH_INFO)
    content_type = MIME_TYPES[file_extension]

    file_exists = False

    if request:
        request_array_key_value = request.split("&")
        
        for element in request_array_key_value:
            request_key_value = element.split("=")
            key = request_key_value[0]
            value = request_key_value[1]
        
    if os.path.isfile(PATH_INFO): # test if requested file exists
        file_exists = True
    elif os.path.isfile("."+PATH_INFO): # test if requested file just simply needs a ./file to be found
        file_exists = True
        PATH_INFO = "."+PATH_INFO
    elif os.path.isdir(PATH_INFO):
        response = str.encode(f"404 {PATH_INFO} is a directory")
        status = "404 Not Found"
        print(status+" "+str(response)) # make error show in the logs
    else:
        response_string = f"404 {PATH_INFO} not found"
        response = str.encode(response_string)
        # response = b"404 {PATH_INFO} not found"
        status = "404 Not Found"
        print(status+" "+response_string) # make error show in the logs

    if file_exists:
        if file_extension == ".py":
            process = subprocess.Popen(['python', PATH_INFO], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            # Wait for the process to finish and get its output and error
            output, error = process.communicate()
            output_str = output.decode('utf-8') # from bytes to strings
            # error_str = error.decode('utf-8') # from bytes to strings
            # if error_str: print("Error:", error_str)
            response = output_str.encode()
            status = "200 OK"
            print(str(status)+" webserver will send file: "+PATH_INFO) # make error show in the logs
        else:
            with open(PATH_INFO,"r") as f: # it is a static file, read and return contents
                response = f.read().encode()
                status = "200 OK"
                print(str(status)+" webserver will send file: "+PATH_INFO) # make error show in the logs

    # how to process data send via web form  
    if PATH_INFO == "/":
        if method == "POST":
            # getting wsgi.input obj
            input_obj = environ["wsgi.input"]
            # length of body
            input_length = int(environ["CONTENT_LENGTH"])
            # getting body of wsgi.input obj
            # decoding to string
            body = input_obj.read(input_length).decode()

            # parsing body of form
            data = urllib.parse.parse_qs(body, keep_blank_values=True)
            # data of body in format
            req = {
                "name" : data["name"][0],
                "email" : data["email"][0],
                "type" : data["f-type"][0],
                "content" : data["feedback"][0]
            }
            # adding to submission
            forms_data.append(req)
            
            response = b"Your feedback submitted successfully."
            status = "200 OK"
    elif PATH_INFO == "/forms":
        # if /forms PATH_INFO
        # converting to JSON data
        response = json.dumps(forms_data).encode()
        status = "200 OK"
        # changing content-type
        content_type = "application/json"

    # response headers
    headers = [
        ("Content-Type",content_type),
        ("Content-Length",str(len(response)))
    ]

    start_response(status, headers)
    return [response]

def start():
    print("...starting webserver")
    # if __name__ == "__main__": # not sure what this if was for, it prevents running the webserver as an included.py file from another.py file
    w_s = wsgiref.simple_server.make_server(host="localhost",port=SETTING_PORT,app=application)
    w_s.serve_forever()
    
# actually start the webserver
# possibility to autostart browser and point to a site
# import webbrowser
# webbrowser.open('http://localhost:8021')
webserver_wsgi.start()